"use strict";
const PLUGIN_NAME = "gulp-gwd";
const through = require("through2");
const gutil = require("gulp-util");
const PluginError = gutil.PluginError;
const fs = require("graceful-fs")
const path = require("path");
const md5File = require("md5-file")

var assets = [];
var assetsRegExp = "";
var checkedAssets = [];
var inlinedAssets = [];
var folder = "";
var logFunc = function(str){ console.log(str) };
var isHashEnabled = false;

module.exports = function(options) {
	var indexExists = false;
	var htmlFiles = [];
	var removedFiles = [], inlinedFiles = [];
	
	options = options || {};
	options.step = options.step || "";
	options.ignored = options.ignored || [];
	options.required = options.required || [];
	
	if (options.logFunc !== null && typeof options.logFunc === "function"){
		logFunc = options.logFunc;
	}
	
	if (options.isHashEnabled !== null && typeof options.isHashEnabled === "boolean"){
		isHashEnabled = options.isHashEnabled;
	}

	function getFiles(file, encoding, callback) {
		if (file.isNull()) { callback(null, file); return; }
		if (file.isStream()) { callback(new PluginError(PLUGIN_NAME, "Streaming not supported")); return; }
		try {
			var filepath = String(file.history);
			var extname = path.extname(filepath);
			var filename = path.basename(filepath);
			
			if (extname === ".html") { 
				htmlFiles.push(String(filepath));
				if (filename === "index.html"){
					checkedAssets.push(String(filepath));
					folder = path.dirname(filepath);
					indexExists = true;
				}
			} else {
				if (!options.ignored.includes(filename)){
					assets.push(String(filepath));
				} else {
					checkedAssets.push(String(filepath));
				}
			}
			this.push(file);
			
		} catch (err) {
			this.emit("error", new PluginError(PLUGIN_NAME, err));
		}
		callback();
	}
	function renameHtml(callback) {
		if (indexExists === true) { 
			// Index уже есть, folder определен
		} else if (htmlFiles.length === 1){
			// Html файл есть, но его нужно переименовать
			var oldPath = path.normalize(htmlFiles[0]);
			folder = path.dirname(oldPath);
			var newPath = path.join(folder, "index.html"); 
			if (fs.existsSync(oldPath)){
				fs.createReadStream(oldPath).pipe(fs.createWriteStream(newPath));
				checkedAssets.push(String(newPath));
			} else {
				this.emit("error", new PluginError(PLUGIN_NAME, "HTML file error"));
			}
			
		} else if (htmlFiles.length > 1){
			this.emit("error", new PluginError(PLUGIN_NAME, "Found " + htmlFiles.length + " .html files, just 1 needed"));
		} else {
			this.emit("error", new PluginError(PLUGIN_NAME, "HTML file is missing"));
		}
		// Добавление style.css (имитация)
		assets.push(String(path.join(folder, "style.css"))); 
		
		// Создание regexp строки
		var regAssets = [];
		for (var i = 0; i < assets.length; i++) {
			regAssets[i] = path.relative(folder, assets[i]).replace(/(\\|\/)/g, unescape("[\\/|\\\\]"));
		}
		
		assetsRegExp = new RegExp(unescape("[\"\'\(][.]?[\\/|\\\\]?") + "(" + regAssets.join("|") + ")[\"\'\)]", "g");
		//console.log(assetsRegExp);
		
		callback();
	};
	
	function replacer(str, group, offset) {
		if (group.indexOf("data:") === 0 || path.isAbsolute(group) === true) return str;
		var newName = String(path.basename(group));
		var oldPath = path.join(folder, group);
		var extname = path.extname(group);
		
		var md5 = (extname === ".css" || extname === ".js") ? md5File.sync(String(oldPath)).substr(0, 5) : "";
		
		if (group !== newName){
			var newPath = path.join(folder, newName);
			
			if (!checkedAssets.includes(newPath)){
				checkedAssets.push(newPath);
				if (fs.existsSync(oldPath) && fs.statSync(oldPath).isFile() === true) {
					fs.renameSync(oldPath, newPath);
				}
			}
			return str.replace(group, addHash(newName, md5));
			
		} else {
			if (!checkedAssets.includes(oldPath)){
				checkedAssets.push(oldPath);
			}
			return str.replace(group, addHash(group, md5));
		}
	}
	function addHash(str, hash) {
		if (isHashEnabled === false) return str;
		if (hash !== "" && str.indexOf("?") === -1){
			return str + "?" + hash;
		} 
		return str;
	}
	function checkAssets(file, encoding, callback) {
		if (file.isNull()) { callback(null, file); return; }
		if (file.isStream()) { callback(new PluginError(PLUGIN_NAME, "Streaming not supported")); return; }
		try {
			var filepath = String(file.history);
			var filename = path.basename(filepath);
			if (!checkedAssets.includes(filepath)){ callback(); return; }
			
			// Производим поиск assets, перенос файлов и переименования
			var data = file.contents.toString();
			data = data.replace(assetsRegExp, replacer);
			file.contents = new Buffer(data);
			this.push(file);
			
		} catch (err) {
			this.emit("error", new PluginError(PLUGIN_NAME, err));
		}
		callback();
	}
	
	function output_filter(file, encoding, callback) {
		if (file.isNull()) { callback(null, file); return; }
		if (file.isStream()) { callback(new PluginError(PLUGIN_NAME, "Streaming not supported")); return; }
		try {
			var filepath = String(file.history);
			var filename = path.basename(filepath);
			
			if (inlinedAssets.includes(String(filepath))){ 
				inlinedFiles.push(filename);
			} else if (checkedAssets.includes(String(filepath))){ 
				this.push(file);
			} else {
				var extname = path.extname(filename);
				if (extname !== ".html" && /\\greensock\\/i.test(filepath) === false){
					removedFiles.push(filename);
				}
			}
		} catch (err) {
			this.emit("error", new PluginError(PLUGIN_NAME, err));
		}
		
		callback();
	}
	function output_report(callback) {
		if (inlinedFiles.length > 0){
			logFunc('<b style="font-size:15px;">&#x27A4;</b>&nbsp; Inlined files: <b>' + inlinedFiles.join(" &middot ") + '</b><br>');
		}
		if (removedFiles.length > 0){
			logFunc('<b style="font-size:18px;line-height:15px;">&#x2716;</b>&nbsp; Removed files: <b>' + removedFiles.join(" &middot ") + '</b><br>');
		}
		
		if (options.required.length > 0) {
			for (var i = 0; i < options.required.length; i++) {
				var req = options.required[i];
				var reqFiles = req[0]; 
				var found = false;
				for (var j = 0; j < reqFiles.length; j++) {
					if (checkedAssets.includes(path.join(folder, reqFiles[j]))){
						found = true; break;
					}
				}
				if (found === false){
					logFunc('<br><b style="color:red;font-size:30px;line-height:0;">&#x2718;</b> <b style="color:red">' + req[1] + ' is missing</b><br>');
				} else {
					logFunc('<br><b style="color:green;font-size:30px;line-height:0;">&#x2713;</b> ' + req[1] + ' is found</b><br>');
				}
			}
		}
		
		callback();
	};
	
	if (options.step === "init"){
		return through.obj(getFiles, renameHtml);
	} else if (options.step === "assets"){
		return through.obj(checkAssets);
	}else if (options.step === "output"){
		return through.obj(output_filter, output_report);
	}
};

module.exports.inlinedFiles = function (arr) {
	inlinedAssets = arr.map(function(relativePath) {
		return path.join(folder, relativePath);
	});
	
	/* Удаляем заинлайненные файлы из checkedAssets
	for (var relativePath in arr) {
		var index = checkedAssets.indexOf(path.join(folder, arr[relativePath]));
		if (index > -1) {
			checkedAssets.splice(index, 1);
		}
	}*/
}